/*
    Copyright (c) 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <sstream>
#include <unistd.h>
#include <list>

#include "database.h"
#include "create_database.h"
#include "config_manager.h"
#include "util.h"

// quote
#define Q(data) "\"" << data << "\""

#define TASKDATA_TRANSMISSION_QUALITY Q("taskdata_test_transmission_quality")

namespace dSA
{

Database* Database::m_instance = NULL;

Database* Database::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new Database();
        if (m_instance == NULL)
        {
            throw _Exception("could not create Database class");
        }
    }

    return m_instance;
}

Database* Database::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("Database not initialized");
    }

    return m_instance;
}

Database::Database()
{
    ConfigManager *cfg = ConfigManager::getInstance();

    bool exists = fileutil::exists(cfg->getOption(CFG_DATABASE_FILE));

    m_sqlite3 = new SQLite3(cfg->getOption(CFG_DATABASE_FILE));
    if (m_sqlite3 == NULL)
    {
        throw _Exception("Failed to allocate database");
    }
    if (!exists)
    {
        try
        {
            fprintf(stderr, "Creating database...\n");
            m_sqlite3->exec(CREATE_DB_V1);
        }
        catch (Exception ex)
        {
            delete m_sqlite3;
            throw _Exception("Could not create database, " + ex.getMessage());
        }
    }

    try
    {
        m_version = util::str2int(getConfig("db_version"));
    }
    catch (Exception ex)
    {
        delete m_sqlite3;
        throw _Exception("Could not determine database layout version, " +
                         ex.getMessage());
    }
}

// internal settings are saved in a table as key:value pairs
std::string Database::getConfig(std::string key)
{
    std::ostringstream qstr;

    qstr << "SELECT " << Q("value") << " FROM " << Q("dsa_config");
    qstr << " WHERE " << Q("key") << " = ";
    qstr << m_sqlite3->escape(key);
    qstr << " LIMIT 1;";

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    // we expect exactly one row (or none if the key does not exist)
    if (q->size() <  1)
    {
        throw _Exception("setting " + key + " was not found in the database");
    }

    // in response we should get exactly one row, containing the value
    boost::shared_ptr<SQLite3::row_result> row = q->at(0);

    if (row->size() < 1)
    {
        throw _Exception("setting " + key + " was not found in the database");
    }

    boost::shared_ptr<SQLite3::cell> cell = row->at(0);
    return cell->data;
}

void Database::setConfig(std::string key, std::string value)
{
    std::ostringstream qstr;

    qstr << "INSERT OR REPLACE INTO " << Q("dsa_config") << " (" <<  Q("key");
    qstr << ", " << Q("value") << ") VALUES (" << m_sqlite3->escape(key);
    qstr << ", " << m_sqlite3->escape(value) << ")";
    m_sqlite3->exec(qstr.str());
}

long long int Database::saveTask(boost::shared_ptr<const dSLongTask> task)
{
    std::ostringstream qstr;

    boost::mutex::scoped_lock(m_mutex);
    qstr << "INSERT INTO " << Q("tasks") << " (" << Q("status") << ",";
    qstr << Q("queued") << "," << Q("type") << "," << Q("parameters") << ") ";
    qstr << "VALUES (" << task->getTaskStatus() << ",";
    qstr << "strftime('%s', 'now'),"<< task->getTaskType() << ",";
    qstr << m_sqlite3->escape(task->getParameters()) << ");";
    m_sqlite3->exec(qstr.str());
    return m_sqlite3->getLastInsertedRowId();
}

void Database::removeTask(long long int task_id)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("tasks") << " WHERE " << Q("task_id") << " = ";
    qstr << task_id << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::removeTasksByStatus(task_status status)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("tasks") << " WHERE " << Q("status") << " = ";
    qstr << status << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::updateTaskStatus(long long int task_id, task_status status)
{
    std::ostringstream qstr;

    qstr << "UPDATE " << Q("tasks") << " SET " << Q("status") << " = ";
    qstr << status;
    if (status == Running)
    {
        qstr << "," << Q("started") << " = " << "strftime('%s', 'now')";
    }
    else if (status == Finished)
    {
        qstr << "," << Q("finished") << " = " << "strftime('%s', 'now')";
    }

    qstr << " WHERE " << Q("task_id") << " = " << task_id << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::updateTaskStatus(task_status old_status, task_status new_status)
{
    std::ostringstream qstr;
    qstr << "UPDATE " << Q("tasks") << " SET " << Q("status") << " = ";
    qstr << new_status;
    qstr << " WHERE " << Q("status") << " = " << old_status << ";";
    m_sqlite3->exec(qstr.str());
}

task_type Database::getTaskType(long long int task_id)
{
    std::ostringstream qstr;
    qstr << "SELECT " << Q("type") << " FROM " << Q("tasks") << " WHERE ";
    qstr << Q("task_id") << " = " << task_id << " LIMIT 1;";
    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    if (q->size() == 0)
    {
        throw _Exception("Task with id " + util::str(task_id) + " not found");
    }

    boost::shared_ptr<SQLite3::row_result> rr = q->at(0);
    for (size_t i = 0; i < rr->size(); i++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(i);
        if (cell->name == "type")
        {
            return (task_type)util::str2int(cell->data);
        }
    }

    throw _Exception("Could not find task type information");
}

boost::shared_ptr<dSLongTask> Database::getTask(long long int task_id)
{
    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("tasks") << " WHERE ";
    qstr << Q("task_id") << " = " << task_id << " LIMIT 1;";
    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    if (q->size() == 0)
    {
        throw _Exception("Task with id " + util::str(task_id) + " not found");
    }

    return createTaskFromRow(q->at(0));
}

boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > >
                                 Database::getTasksByStatus(task_status status)
{
    boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > >
        tasks(new std::list<boost::shared_ptr<dSLongTask> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << Q("tasks") << " WHERE " << Q("status") << " = ";
    qstr << status << ";";
    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        try
        {
            tasks->push_back(createTaskFromRow(q->at(i)));
        }
        catch (Exception ex)
        {
            fprintf(stderr, "Caught exception while parsing saved tasks, "
                    "%s\n", ex.getMessage().c_str());
            continue;
        }
    }

    return tasks;
}

void Database::saveTransmissionQuality(long long int task_id,
                                       std::string device_dsid,
                                       uint16_t upstream, uint8_t downstream,
                                       int error_code)
{
    std::ostringstream qstr;

    qstr << "INSERT INTO " << TASKDATA_TRANSMISSION_QUALITY << " (";
    qstr << Q("dsid") << ",";
    qstr << Q("upstream") << ",";
    qstr << Q("downstream") << ",";
    qstr << Q("timestamp") << ",";
    qstr << Q("errorcode") << ",";
    qstr << Q("task_id");
    qstr << ") ";

    qstr << "VALUES (";
    qstr << m_sqlite3->escape(device_dsid) << ",";
    qstr << upstream << ",";
    qstr << (int)downstream << ",";
    qstr << "strftime('%s', 'now'),";
    qstr << error_code << ",";
    qstr << task_id;
    qstr << ");";

    m_sqlite3->exec(qstr.str());
}

boost::shared_ptr<Database::task_result_summary>
                          Database::getTaskResultSummary(long long int task_id)
{
    boost::shared_ptr<Database::task_result_summary> result(
            new Database::task_result_summary);

    boost::shared_ptr<result_data> data(new result_data);

    result->task = getTask(task_id);
    result->data = data;

    std::ostringstream qstr;

    boost::shared_ptr<SQLite3::query_result> q;

    if (result->task->getTaskType() == TestTransmissionQuality)
    {
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
        qstr << "WHERE " << Q("errorcode") << " != 0 AND ";
        qstr << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["errors"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
        qstr << "WHERE " << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["tests"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT COUNT(DISTINCT " << Q("dsid") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["devices"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MAX( " << Q("upstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["maxUpstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MIN( " << Q("upstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["minUpstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT AVG( " << Q("upstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["avgUpstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MAX( " << Q("downstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["maxDownstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MIN( " << Q("downstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["minDownstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT AVG( " << Q("downstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["avgDownstream"] = singleValueFromQuery(q);
    }
    else
    {
        throw _Exception("could not load task data, unknown task");
    }

    return result;
}

boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                 Database::loadTasks(long long int limit,
                                                     long long int offset)
{
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
        tasks(new std::list<boost::shared_ptr<const dSLongTaskInfo> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << Q("tasks");
    if (limit == 0)
    {
        limit = -1;
    }

    qstr << " LIMIT " << limit;

    if (offset > 0)
    {
        qstr << " OFFSET " << offset;
    }

    qstr << ";";

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        try
        {
            tasks->push_back(createTaskFromRow(q->at(i)));
        }
        catch (Exception ex)
        {
            fprintf(stderr, "Caught exception while parsing saved tasks, "
                    "%s\n", ex.getMessage().c_str());
            continue;
        }
    }

    return tasks;
}
boost::shared_ptr<dSLongTask> Database::createTaskFromRow(
                                     boost::shared_ptr<SQLite3::row_result> rr)
{
    task_type tt = UnknownTask;
    task_status ts = UnknownStatus;
    webutil::query params;
    long long int task_id = 0;
    time_t started = 0;
    time_t finished = 0;
    time_t queued = 0;

    for (size_t j = 0; j < rr->size(); j++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(j);
        if (cell->name == "type")
        {
            tt = (task_type)util::str2int(cell->data);
        }
        else if (cell->name == "status")
        {
            ts = (task_status)util::str2int(cell->data);
        }
        else if (cell->name == "parameters")
        {
            webutil::parse_query(cell->data, params);
        }
        else if (cell->name == "task_id")
        {
            task_id = util::str2llint(cell->data);
        }
        else if (cell->name == "started")
        {
            if (!cell->data.empty())
            {
                started = util::str2time_t(cell->data);
            }
        }
        else if (cell->name == "queued")
        {
            if (!cell->data.empty())
            {
                queued = util::str2time_t(cell->data);
            }
        }
        else if (cell->name == "finished")
        {
            if (!cell->data.empty())
            {
                finished = util::str2time_t(cell->data);
            }
        }
    }

    if (tt == TestTransmissionQuality)
    {
        if (params.size() < 2)
        {
            throw _Exception("not enough parameters loaded for "
                             "test transmission quality task");
        }

        boost::shared_ptr<TestTranmissionQualityTask> task(
            new TestTranmissionQualityTask(
                util::str2int(params["checksPerDevice"]),
                util::str2dsid(params["dsm"])));
        task->setTaskStatus(ts);
        task->setTaskID(task_id);
        task->setStartedTime(started);
        task->setFinishedTime(finished);
        task->setQueuedTime(queued);
        return task;
    }

    throw _Exception("failed to identify task, unknown task type");
}

boost::shared_ptr<Database::field> Database::singleValueFromQuery(
                                    boost::shared_ptr<SQLite3::query_result> q)
{
    if (q->size() == 0)
    {
        throw _Exception("failed to retrieve result");
    }

    boost::shared_ptr<SQLite3::row_result> rr = q->at(0);

    if (rr->size() == 0)
    {
        throw _Exception("failed to retrieve result");
    }
    boost::shared_ptr<SQLite3::cell> cell = rr->at(0);

    boost::shared_ptr<Database::field> d(
            new Database::field(cell->data, cell->type));
    return d;
}

long long int Database::transmissionQualityResultCount(long long int task_id)
{
    std::ostringstream qstr;
    qstr << "SELECT COUNT(*) FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
    qstr << "WHERE " << Q("task_id") << " = " << task_id << ";";

    boost::shared_ptr<SQLite3::query_result> q;
    q = m_sqlite3->query(qstr.str());
    boost::shared_ptr<Database::field> f = singleValueFromQuery(q);

    if (f->second != SQLITE_INTEGER)
    {
        throw _Exception("unexpected datatype");
    }
    return util::str2llint(f->first);
}


boost::shared_ptr<Database::transmission_quality_device_record>
    Database::TQDevRecordFromRow(boost::shared_ptr<SQLite3::row_result> rr)
{
    boost::shared_ptr<transmission_quality_device_record>
        rec(new Database::transmission_quality_device_record);

    for (size_t j = 0; j < rr->size(); j++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(j);
        if (cell->name == "dsid")
        {
            rec->dsid = cell->data;
        }
        else if (cell->name == "upstream")
        {
            rec->upstream = util::str2int(cell->data);
        }
        else if (cell->name == "downstream")
        {
            rec->downstream = util::str2int(cell->data);
        }
        else if (cell->name == "timestamp")
        {
            rec->timestamp = util::str2time_t(cell->data);
        }
        else if (cell->name == "errorcode")
        {
            rec->errorcode = util::str2int(cell->data);
        }
    }

    return rec;
}

boost::shared_ptr<std::list<
    boost::shared_ptr<const Database::transmission_quality_device_record> > >
            Database::getTransmissionQualityResults(long long int task_id,
                                                    long long int limit,
                                                    long long int offset)
{
    boost::shared_ptr<std::list<
        boost::shared_ptr<
            const Database::transmission_quality_device_record> > >
                ret(new std::list<
                    boost::shared_ptr<
                        const Database::transmission_quality_device_record> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
    qstr << "WHERE " << Q("task_id") << " = " << task_id;
    if (limit == 0)
    {
        limit = -1;
    }

    qstr << " LIMIT " << limit;

    if (offset > 0)
    {
        qstr << " OFFSET " << offset;
    }

    qstr << ";";

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        try
        {
            ret->push_back(TQDevRecordFromRow(q->at(i)));
        }
        catch (Exception ex)
        {
            fprintf(stderr, "Caught exception while parsing test results, "
                    "%s\n", ex.getMessage().c_str());
            continue;
        }
    }

    return ret;
}

Database::~Database()
{
    if (m_sqlite3)
    {
        delete m_sqlite3;
    }
}

} // namespace
