/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Michael Tross <mt@aizo.ag>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <boost/format.hpp>
#include "hwstats.h"
#include "storage.h"
#include "util.h"

namespace dSA
{

StatisticBase::StatisticBase(uint8_t version)
{
    m_version = version;
    m_index = 0;
    memset(m_data, 0, sizeof(m_data));
    memset(m_valid, 0, sizeof(m_valid));
}

StatisticBase::~StatisticBase() {}

uint8_t StatisticBase::getVersion() const
{
    return m_version;
}

uint8_t StatisticBase::getValue8(uint8_t index) const
{
    return m_data[index];
}

uint16_t StatisticBase::getValue16(uint8_t index) const
{
    uint16_t rval =
            (m_data[index+1] << 8) |
            (m_data[index]);
    return rval;
}

uint32_t StatisticBase::getValue32(uint8_t index) const
{
    uint32_t rval =
            (m_data[index+3] << 24) |
            (m_data[index+2] << 16) |
            (m_data[index+1] << 8) |
            (m_data[index+0]);
    return rval;
}

void StatisticBase::pushValue(uint8_t val)
{
    m_valid[m_index] = true;
    m_data[m_index++] = val;
}

void StatisticBase::pushValue(uint16_t val)
{
    pushValue( (uint8_t) (val & 0xff));
    pushValue( (uint8_t) (val >> 8));
}

void StatisticBase::pushNilValue(uint8_t num)
{
    m_index += num;
}

StatisticDevice::StatisticDevice(uint8_t version) :
        StatisticBase(version)
{

}

boost::shared_ptr<std::list<std::string> > StatisticDevice::getResults()
{
    typedef enum {
        unsignedint,
        signedint,
        dsc11temp,
        zeroxcount,
    } named_value_data_t;
    typedef struct _named_value {
        const char *name;
        const char *desc;
        uint8_t size;
        named_value_data_t type;
    } named_value_t;

    named_value_t sdesc_v1[] = {
            { "Version", "dataset version", 1, unsignedint },
            { "Length", "length of dataset in bytes", 1, unsignedint },
            { "Temperature", "chip temperature", 1, dsc11temp },
            { "Polarity", "polarity of receiver", 1, unsignedint },
            { "Scene", "current scene value", 1, unsignedint },
            { "LastScene", "previous scene value", 1, unsignedint },
            { "OverTempCnt", "overtemperature event counter", 2, unsignedint },
            { "OverCurrCnt", "overcurrent (shortcirc) event counter", 2,
              unsignedint },
            { "OverLoadCnt", "overload event counter", 2, unsignedint },
            { "NetOverLoadCnt", "net overload event counter", 2, unsignedint },
            { "RcMsgLongCnt", "count of transmitted long rev chan messages", 2,
               unsignedint },
            { "RcMsgShortCnt", "Count of transmitted short rev chan messages",
              2, unsignedint },
            { "FcHalfWaves", "Halfwaves", 4, zeroxcount },
            { "FcStartLongCnt", "start-words long", 2, unsignedint },
            { "FcStartShortCnt2", "start-words short", 2, unsignedint },
            { "FcMsgLongCnt", "correct long", 2, unsignedint },
            { "FcMsgShortCnt", "correct short", 2, unsignedint },
            { "FcErr10Cnt",
              "corrected bits/hamming codewords (corrected from 1 to 0)", 2, 
              unsignedint },
            { "FcErr01Cnt",
              "corrected bits/hamming codewords (corrected from 0 to 1)", 2,
              unsignedint },
            { "FcErr2Cnt", "discarded hamming codewords", 2, unsignedint },
            { "FcBitLen[6]",
              "average bit length from within correct messages only excl. "
              "start word (01Beg, 01End, 10Beg, 10End, 11Beg, 11End)", 6,
              unsignedint },
            { "FcBitVar[6]", "variance of bit length from within correct"
              " messages only excl. start word (01Beg, 01End, 10Beg, 10End, "
              "11Beg, 11End)", 12, unsignedint },
            { "FcRecToBeg", "#of receiver timeouts (beginning of hw)", 2,
              unsignedint },
            { "FcRecToEnd", "#of receiver timeouts (end of hw)", 2,
              unsignedint },
            { "FcRecErrBeg", "#of receiver errors (beginning of hw)", 2,
              unsignedint },
            { "FcRecErrEnd", "#of receiver errors (end of hw)", 2,
              unsignedint },
            { "FcBitNoMsgBeg", "#of received pulses (non-zeros) not within "
              "correctly decoded telegrams (beginning of hw)", 2, unsignedint },
            { "FcBitNoMsgEnd", "#of received pulses (non-zeros) not within "
              "correctly decoded telegrams (end of hw)", 2, unsignedint },
            { "FcBitMsgBeg", "#of received pulses (non-zeros) from within "
              "correctly decoded telegrams (beginning of hw)", 2, unsignedint },
            { "FcBitMsgEnd", "#of received pulses (non-zeros) from within "
              "correctly decoded telegrams (end of hw)", 2, unsignedint },
    };

    named_value_t sdesc_v4[] = {
            { "Version", "dataset version", 1, unsignedint },
            { "Length", "length of dataset in bytes", 1, unsignedint },
            { "Temperature", "chip temperature", 1, dsc11temp },
            { "Polarity", "polarity of receiver", 1, unsignedint },
            { "Scene", "current scene value", 1, unsignedint },
            { "LastScene", "previous scene value", 1, unsignedint },
            { "OverTempCnt", "overtemperature event counter", 2, unsignedint },
            { "OverCurrCnt", "overcurrent (shortcirc) event counter", 2,
              unsignedint },
            { "OverLoadCnt", "overload event counter", 2, unsignedint },
            { "NetOverLoadCnt", "net overload event counter", 2, unsignedint },
            { "RcMsgLongCnt", "count of transmitted long upstream messages", 2,
              unsignedint },
            { "RcMsgShortCnt", "count of transmitted short upstream messages",
              2, unsignedint },
            { "RcMsgShortResendCnt",
              "count of repeated short upstream messages", 2, unsignedint },
            { "FcHalfWaves", "Halfwaves", 4, zeroxcount },
            { "FcStartCnt", "start-words", 2, unsignedint },
            { "FcMsgLongCnt", "correct long downstream messages", 2,
              unsignedint },
            { "FcMsgShortCnt", "correct short downstream messages", 2,
              unsignedint },
            { "FcErr10Cnt",
              "corrected bits/hamming codewords (corrected from 1 to 0)", 2,
              unsignedint },
            { "FcErr01Cnt",
              "corrected bits/hamming codewords (corrected from 0 to 1)", 2,
              unsignedint },
            { "FcErr2Cnt", "discarded hamming codewords", 2, unsignedint },
            { "FcBitLenAv1", "average bit length (1st bit)", 1, unsignedint },
            { "FcBitLenAv2", "average bit length (2nd bit)", 1, unsignedint },
            { "FcBitLenMin1", "#minimal bit length (1st bit)", 1, unsignedint },
            { "FcBitLenMin2", "#minimal bit length (2nd bit)", 1, unsignedint },
            { "FcBitLenMax1", "#maximal bit length (1st bit)", 1, unsignedint },
            { "FcBitLenMax2", "#maximal bit length (2nd bit)", 1, unsignedint },
            { "FcBitIvCnt[16]", "#of bits by interval (8x 1st/2nd) "
              "(0..31/32..39/40..47/48..55/56..63/64..71/72..79/80..127)", 32,
              unsignedint },
            { "FcRecTo1", "#of receiver timeouts (1st bit)", 2, unsignedint },
            { "FcRecTo2", "#of receiver timeouts (2nd bit)", 2, unsignedint },
            { "FcRecErr1", "#of receiver errors (1st bit)", 2, unsignedint },
            { "FcRecErr2", "#of receiver errors (2nd bit)", 2, unsignedint },
            { "FcBitNoMsg1", "#of received pulses (non-zeros) not within "
              "correctly decoded telegrams (1st bit)", 2, unsignedint },
            { "FcBitNoMsg2", "#of received pulses (non-zeros) not within "
              "correctly decoded telegrams (2nd bit)", 2, unsignedint },
            { "FcBitMsg1", "#of received pulses (non-zeros) from within "
              "correctly decoded telegrams (1st bit)", 2, unsignedint },
            { "FcBitMsg2", "#of received pulses (non-zeros) from within "
              "correctly decoded telegrams (2nd bit)", 2, unsignedint },
            { "ChipResetCnt", "dSC11 reset event counter", 2, unsignedint },
            { "OutChangedCnt", "Count of output changes (e.g. Scene Call, "
              "local push button)", 2, unsignedint },
            { "FcHalfWavesOn", "#Halfwaves Device turned on", 4, zeroxcount },
            { "LTGlitchCnt", "#of LT-Button glitch events", 1, unsignedint },
            { "LTButtonCnt", "#of LT-Button press event", 1, unsignedint },
            { "EXTGlitchCnt", "#of EXT-Button glitch events", 1, unsignedint },
            { "EXTButtonCnt", "#of EXT-Button press events", 1, unsignedint },
            { "FcDiscLongCnt", "#discarded long downstream messages", 2,
              unsignedint },
            { "FcDiscShortCnt", "#discarded short downstream messages", 2,
              unsignedint },
    };

    boost::shared_ptr<std::list<std::string> > result(
                                                    new std::list<std::string>);
    uint8_t ver = m_data[1];
    uint8_t len = m_data[0];
    int i, i0;
    named_value_t *iter;

    // select right table based on ver
    switch (ver) {
        case 0x01:
        case 0x02:
            iter = sdesc_v1;
            break;
        case 0x03:
        case 0x04:
        default:
            iter= sdesc_v4;
            break;
    }
    len = len;

    // iterate over named value table
    for(i = i0 = 0; (i <= m_index) && (i0 < m_index); i++) {
        if (i - i0 >= iter->size) {

            bool incomplete = false;
            for (int n = 0; n < iter->size; n ++) {
                if (!m_valid[i0+n])
                    incomplete = true;
            }
            if (incomplete) {
                result->push_back(std::string(iter->name) + "=incomplete");
                i0 += iter->size;
                iter ++;
                continue;
            }

            switch (iter->type)
            {
            case unsignedint:
            case signedint:
            default:
                switch (iter->size)
                {
                case 1:
                {
                    uint8_t val = getValue8(i0);
                    result->push_back(std::string(iter->name) + "="
                            + util::hex(val, 2));
                }
                break;
                case 2:
                {
                    uint16_t val = getValue16(i0);
                    result->push_back(std::string(iter->name) + "="
                            + util::hex(val, 4));
                }
                break;
                case 4:
                {
                    uint32_t val = getValue32(i0);
                    result->push_back(std::string(iter->name) + "="
                                + util::hex(val, 8));
                }
                break;
                default:
                {
                    result->push_back(std::string(iter->name) + "="
                            + util::hex(&m_data[i0], iter->size));
                }
                break;
                }

                break;

                case zeroxcount:
                {
                    uint32_t val = getValue32(i0);
                    uint32_t seconds = (val + 50) / 100;
                    uint32_t minutes = seconds / 60; seconds %= 60;
                    uint32_t hours = minutes / 60; minutes %= 60;
                    uint32_t days = hours / 24; hours %= 24;
                    std::string uptime(boost::str(
                        boost::format(
                            "%1%d %2%h %3%m %4%s"
                            ) % days % hours % minutes % seconds ));
                    result->push_back(std::string(iter->name) + "=" +
                                util::hex(val, 8) + " (" + uptime + ")" );
                }
                break;

                case dsc11temp:
                {
                    uint8_t val = getValue8(i0);
                    std::string celsius(boost::str(
                        boost::format("%1% °C, +/-10") % (149 - 0.9 * val)));
                    result->push_back(std::string(iter->name) + "=" +
                            util::hex(val, 2) + " (" + celsius + ")");
                }
                break;
            }

            i0 += iter->size;
            iter ++;
        }
    }

    return result;
}

}
