/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <stdlib.h>
#include <signal.h>
#include <limits.h>

#include "ds-assistant.h"
#include "config_manager.h"
#include "database.h"
#include "storage.h"
#include "webserver.h"
#include "log.h"
#include "sqlite3_wrapper.h"

#if defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)
    #include <getopt.h>
    #define OPTSTR "c:p:w:vh"
#else
    #error Need getopt_long!
#endif

using namespace dSA;

static int g_shutdown_flag = 0;

void signal_handler(int signum)
{
    if ((signum == SIGINT) || (signum == SIGTERM))
    {
        g_shutdown_flag++;
    }
    if (signum == SIGUSR1)
    {
        try
        {
            dS_Assistant::getInstance()->reopenLog();
        }
        catch (Exception ex)
        {
            fprintf(stderr, "%s\n", ex.getMessage().c_str());
        }
    }
    return;
}

static void print_copyright()
{
    printf("\ndigitalSTROM Assistant (dSA) version %s\n\n", VERSION);
    printf("Copyright (c) 2010,2011 aizo ag, Zurich, Switzerland\n");
    printf("digitalSTROM Aassistant is free software, distributed under GNU "
           "GPL\n\n");
}

static void print_usage()
{
    printf("\
\n\
Supported options:\n\
    --config or -c              use specified configuration file\n\
    --port or -p                application webserver port\n\
    --version                   print version information and exit\n\
    --help                      this help message\n\
\n");
}

int main(int argc, char **argv)
{
    int port = 0;
    int opt_index = 0;
    int o;
    char *err = NULL;
    struct sigaction action;
    ConfigManager* cfg = NULL;
    const char* cfg_file = NULL;
    const char* webroot = NULL;
    Storage *storage = NULL;
    dS_Assistant* assistant = NULL;
    Database *database = NULL;
    Webserver* webserver = NULL;

    static struct option long_options[] =
    {
        {"config",          1, 0, 'c'},
        {"port",            1, 0, 'p'},
        {"webroot",         1, 0, 'w'},
        {"version",         0, 0, 'v'},
        {"help",            0, 0, 'h'},
        {0, 0, 0, 0}
    };

    print_copyright();

    while (1)
    {
        o = getopt_long(argc, argv, OPTSTR, long_options, &opt_index);
        if (o == -1) break;

        switch (o)
        {
            case 'p':
                errno = 0;
                port = strtol(optarg, &err, 10);
                if ((port == 0) && (*err))
                {
                    fprintf(stderr, "Invalid application port argument: %s\n",
                            optarg);
                    exit(EXIT_FAILURE);
                }

                if (port > USHRT_MAX)
                {
                    fprintf(stderr, "Invalid application port value %d. "
                            "Maximum allowed port value is %d\n", port,
                            USHRT_MAX);
                    exit(EXIT_FAILURE);
                }
                break;
            case 'c':
                cfg_file = optarg;
                break;
            case 'w':
                webroot = optarg;
                break;
            case 'h':
            case '?':
            default:
                print_usage();
                exit(EXIT_FAILURE);
        }
    }

    memset(&action, 0, sizeof(action));
    action.sa_handler = signal_handler;
    action.sa_flags = 0;
    sigfillset(&action.sa_mask);

    if (sigaction(SIGINT, &action, NULL) < 0)
    {
        fprintf(stderr, "Could not register SIGINT handler!\n");
        return EXIT_FAILURE;
    }

    if (sigaction(SIGTERM, &action, NULL) < 0)
    {
        fprintf(stderr, "Could not register SIGTERM handler!\n");
        return EXIT_FAILURE;
    }

    if (sigaction(SIGUSR1, &action, NULL) < 0)
    {
        fprintf(stderr, "Could not register SIGUSR1 handler!\n");
        return EXIT_FAILURE;
    }

    try
    {
        cfg = ConfigManager::createInstance(cfg_file, webroot, port);
        database = Database::createInstance();
        storage = Storage::getInstance();
        assistant = dS_Assistant::createInstance();
        if (assistant == NULL)
        {
            throw _Exception("Could not initialize dS Assistant");
        }

        assistant->run();

        webserver = new Webserver(cfg->getIntOption(CFG_WEBSERVER_PORT),
                                  cfg->getOption(CFG_WEBSERVER_DOCUMENT_ROOT));

        while (!g_shutdown_flag)
        {
            pause();
        }

        throw _Exception("Shutting down...");
    }
    catch (Exception ex)
    {
        fprintf(stderr, "%s\n", ex.getMessage().c_str());
        FILE_LOG(logERROR) << ex.getMessage();
        if (webserver)
        {
            delete webserver;
            webserver = NULL;
        }

        if (assistant)
        {
            delete assistant;
            assistant = NULL;
        }

        if (storage)
        {
            delete storage;
            storage = NULL;
        }

        if (database)
        {
            delete database;
            database = NULL;
        }
        if (cfg)
        {
            delete cfg;
            cfg = NULL;
        }
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
