/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __TASKS_H__
#define __TASKS_H__

#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include <boost/thread/mutex.hpp>
#include <string>
#include <time.h>

#include "dsmapi-wrapper.h"
#include "hwdesc.h"

namespace dSA
{

class dS_Assistant;

// these values are saved in the database, so make sure to only add
typedef enum
{
    UnknownTask             = 0,
    TestTransmissionQuality = 1,
    TestDevicePresence      = 2,
    TestDeviceStatistic     = 3,
} task_type;

typedef enum
{
    UnknownStatus       = 0,
    Queued              = 1,
    Running             = 2,
    Finished            = 3,
    Cancelled           = 4
} task_status;

class dSTaskInfo
{
public:
    dSTaskInfo();
    void setDescription(std::string description);
    const std::string getName() const;
    const std::string getDescription() const;
    bool isValid() const;
    void invalidate();
    bool isCancellable() const;

    virtual ~dSTaskInfo();

protected:
    std::string m_name;
    std::string m_description;
    bool m_valid;
    bool m_cancellable;
};

class dSLongTaskInfo : public dSTaskInfo
{
public:
    dSLongTaskInfo();
    void setTaskID(long long int task_id);
    long long int getTaskID() const;
    void setParentTaskID(long long int task_id);
    long long int getParentTaskID() const;
    unsigned char getProgress() const;
    std::string getStatusInfo() const;
    std::string getParameters() const;
    void setTaskStatus(task_status status);
    task_status getTaskStatus() const;
    std::string getTaskStatusText() const;
    virtual task_type getTaskType() const = 0;
    time_t getStartedTime() const;
    void setStartedTime(time_t t);
    time_t getFinishedTime() const;
    void setFinishedTime(time_t t);
    time_t getQueuedTime() const;
    void setQueuedTime(time_t t);

protected:
    long long int m_task_id;
    long long int m_parent_task_id;
    unsigned char m_progress;
    std::string m_status_info;
    task_status m_task_status;
    mutable boost::mutex m_status_mutex;
    std::string m_parameters;
    time_t m_started;
    time_t m_finished;
    time_t m_queued;
};

class dSTask : public dSTaskInfo
{
public:
    dSTask();
    virtual void run(dS_Assistant *) = 0;
};

class dSLongTask : public dSLongTaskInfo
{
public:
    dSLongTask();
    virtual void run(dS_Assistant *) = 0;
};

class GetBusMembersTask : public dSTask
{
public:
    GetBusMembersTask(bool async = true);
    virtual void run(dS_Assistant *);
private:
    bool m_async;
};

class GetDSMInfoTask : public dSTask
{
public:
    GetDSMInfoTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class GetDSMZonesTask : public dSTask
{
public:
    GetDSMZonesTask(dsid_t dsid);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
};

class GetDSMZoneInfoTask : public dSTask
{
public:
    GetDSMZoneInfoTask(dsid_t dsid, uint8_t zone_index);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint8_t m_zone_index;
};

class ScanDSMTask : public dSTask
{
public:
    ScanDSMTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class StartupDSMTask : public dSTask
{
public:
    StartupDSMTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class SetDSMApartmentIdTask : public dSTask
{
public:
    SetDSMApartmentIdTask(dsid_t dsid, int apartment_id, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    int m_apartment_id;
    bool m_async;
};

class CheckDSMTask : public dSTask
{
public:
    CheckDSMTask(dsid_t dsid, bool async = true);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    bool m_async;
};

class GetAllDevicesTask : public dSTask
{
public:
    GetAllDevicesTask(dsid_t dsid);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
};

class GetActiveDevicesInZoneTask : public dSTask
{
public:
    GetActiveDevicesInZoneTask(dsid_t dsid, uint16_t zone_id);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint16_t m_zone_id;
};

class GetDeviceInfoTask : public dSTask
{
public:
    GetDeviceInfoTask(dsid_t dsm_dsid, uint16_t device_id);
    virtual void run(dS_Assistant *);
private:
    dsid_t m_dsid;
    uint16_t m_device_id;

};

class TestTranmissionQualityTask : public dSLongTask
{
public:
    TestTranmissionQualityTask(int checksPerDevice, dsid_t dsm_dsid);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    int m_checksPerDevice;
    dsid_t m_dsm_dsid;
};

class TestDevicePresenceTask : public dSLongTask, public dSMApiSelectiveListener
{
public:
    TestDevicePresenceTask(dsid_t dsm_dsid, int interval_time);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
    virtual void callback(boost::shared_ptr<CBData> callback_data);

    typedef enum
    {
      unkown,
      scheduled,
      sent,
      replied,
      removed,
      discovered
    } ping_state_t;

    class DeviceP : public Device
    {
    public:
        DeviceP(boost::shared_ptr<const Device> &parent);
        void setPingState(ping_state_t state) { m_ping_state = state; }
        ping_state_t getPingState() const     { return m_ping_state; };
    private:
        ping_state_t m_ping_state;
    };

    class MeterP : public dSM
    {
    public:
        MeterP(boost::shared_ptr<const dSM> &parent);
        void resetPingState();
        void getPingStatus(uint16_t* tx_count, uint16_t* rx_count);
        void signalPingSent(uint16_t device_id);
        void signalPingReply(uint16_t device_id);

        void setSelectionA(uint16_t data) { m_a_selection = data; }
        uint16_t getSelectionA() const    { return m_a_selection; };
        void setSelectionB(uint16_t data) { m_b_selection = data; };
        uint16_t getSelectionB() const    { return m_b_selection; };
        void setSelectionD(uint32_t data) { m_d_selection = data; };
        uint32_t getSelectionD() const    { return m_d_selection; };

    private:
        uint16_t m_count_ping_sent;
        uint16_t m_count_ping_received;
        uint16_t m_a_selection;
        uint16_t m_b_selection;
        uint32_t m_d_selection;
    };

    boost::shared_ptr<MeterP> getDSM(dsid_t dsmid);
    boost::shared_ptr<DeviceP> getDevice(dsid_t dsmid, uint16_t devindex);
    boost::shared_ptr<DeviceP> getDevice(dsid_t dsmid, dsid_t devid);

private:
    dsid_t m_dsm_dsid;
    int m_interval_time;

    boost::shared_ptr<std::list<boost::shared_ptr<DeviceP> > > m_devs;
    boost::shared_ptr<std::list<boost::shared_ptr<MeterP> > > m_dsms;
};

class TestDeviceStatisticTask : public dSLongTask
{
public:
    TestDeviceStatisticTask(dsid_t dsm_dsid, dsid_t dev_dsid);
    virtual void run(dS_Assistant *);
    virtual task_type getTaskType() const;
private:
    dsid_t m_dsm_dsid;
    dsid_t m_dev_dsid;
};

}

#endif//__TASKS_H__
